/* NVTV Chrontel TV-I2C access -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id$
 *
 * Contents:
 *
 * Routines to access the Chrontel chip registers via the I2C bus.
 *
 * Model2: CH7009, CH7010 (and maybe CH7011, CH7012)
 *
 */

#include "xf86i2c.h"
#include "xfree.h"
#include "debug.h"
#include "bitmask.h"
#include "tv_i2c.h"
#include "tv_ch.h"

/* Registers */

dm_ir = dmr_ir
dm_vs = dmr_vs
dm_sr = dmr_sr

ff_yffnt = ffr_fy
ff_yfft  = fft_ft
ff_cff   = fft_fc

vbw_cbw
vbw_ysv
vbw_ycv

te !! new

civc = civh

/* -------- Chrontel -------- */

/* Constant chrontel model 2 register values
 *
 * -  01-6   vof   = 0 (composite and S-Video)
 * -  02-6   cfrb  = 1 (default, lock to vertical) (FIXME extra field)
 * -  02-7   vbid  = 1 (default, minimum flickering during vertical blanking)
 */

void TVChSetRegs (TVEncoderObj *this, TVEncoderRegs *r, TVState state)
{
  register I2CDevPtr dev = this->dev;
  I2CByte *m;
  int i;

  DPRINTF ("tv regs ch\n");
#ifdef FAKE_I2C
  FPRINTF ("\ni2c [");
#endif
  m = TVChMacroRegs (r);
  TVWriteBus (dev, 0x80|0x00, SetBitField(r->ch.dmr_ir,2:0,7:5)
		            | SetBitField(r->ch.dmr_vs,1:0,4:3)
		            | SetBitField(r->ch.dmr_sr,2:0,2:0));
  TVWriteBus (dev, 0x80|0x01, m[0x01] 
	                    | SetBitField(r->ch.ffr_fc,1:0,5:4)
 	                    | SetBitField(r->ch.ffr_ft,1:0,3:2)
		            | SetBitField(r->ch.ffr_fy,1:0,1:0));
  TVWriteBus (dev, 0x80|0x02, SetBit(7) /* vbid */
	                    | SetBitFlag(r->ch.flags,CH_FLAG_CFRB,6)
	                    | SetBitFlag(r->ch.flags,CH_FLAG_CVBW,5)
	                    | SetBitField(r->ch.vbw_cbw,  0:0,4:4)
	                    | SetBitField(r->ch.vbw_ysv,  1:0,3:2)
			    | SetBitField(r->ch.vbw_ycv,  1:0,1:0));
  TVWriteBus (dev, 0x80|0x03, m[0x03] 
	                    | SetBitField(r->ch.sav,8:8,5:5)
		            | SetBitField(r->ch.hpr,8:8,4:4)
		            | SetBitField(r->ch.vpr,8:8,3:3)
	                    | SetBitField(r->ch.te, 2:0,2:0));
  TVWriteBus (dev, 0x80|0x04, Set8Bits(r->ch.sav));
  TVWriteBus (dev, 0x80|0x05, Set8Bits(r->ch.hpr));
  TVWriteBus (dev, 0x80|0x06, Set8Bits(r->ch.vpr));
  TVWriteBus (dev, 0x80|0x07, Set8Bits(r->ch.blr));
  TVWriteBus (dev, 0x80|0x08, m[0x08] 
	                    | SetBitField(r->ch.ce,2:0,2:0));
  TVWriteBus (dev, 0x80|0x09, 0x80 /* mem=2, ibi=0, pllcpi=0 */
	                    | SetBitField(r->ch.pll_n,9:8,4:3)
		            | SetBitField(r->ch.pll_m,8:8,2:2)
	                    | SetBitField(r->ch.pllcap,0:0,0:0));
  TVWriteBus (dev, 0x80|0x0a, Set8Bits(r->ch.pll_m));
  TVWriteBus (dev, 0x80|0x0b, Set8Bits(r->ch.pll_n));
  TVWriteBus (dev, 0x80|0x0c, SetBitField(r->ch.fsci,31:24,7:0));
  TVWriteBus (dev, 0x80|0x0d, SetBitField(r->ch.fsci,23:16,7:0));
  TVWriteBus (dev, 0x80|0x0e, SetBitField(r->ch.fsci,15: 8,7:0));
  TVWriteBus (dev, 0x80|0x0f, SetBitField(r->ch.fsci, 7: 0,7:0));
  TVWriteBus (dev, 0x80|0x10, m[0x10] /* civc_paln = 0 */
	                    | SetBitField(r->ch.civh,1:0,3:2)
	                    | SetBitFlag(r->ch.flags,CH_FLAG_ACIV,0));
  /* 0x11-0x13 civ -- read only */
  /* 0x14-0x1b resverd */
  TVWriteBus (dev, 0x80|0x1c, m[0x1c] 
	                    | 0x00 /* FIXME m/s etc. hwconfig */ );
  TVWriteBus (dev, 0x80|0x1d, m[0x1d] 
	                    | 0x08 /* xcmd=8 */ );
  /* FIXME 0x1e hwconfig */
  /* FIXME 0x1f port */
  /* 0x20 reserved */
  /* FIXME 0x21 dac control */
  /* 0x21      bco -- not used */
  /* 0x22-0x30 reserved */
  /* 0x31-0x37 dvi -- not used */
  /* 0x38-0x47 reserved */
  /* 0x48      test pattern */
  /* 0x49      power management */
  /* 0x4a-0x4b version */
  /* 0x4c-0x4f reserved */

#ifdef FAKE_I2C
  FPRINTF ("]\n");
#endif
}
 


/*

APS0: 15
APS1-3: 15 18 19/1f 1a 1b 28 29 2a/0f 2b/3f 2c 2d 2e/3f 2f/3f 
        40 41 42/7f 43 44/7f 45

Mode: 16/7f 17 38/7f 39/0f 3a 3b 3c/7f 3d/7f 3e 3f

APS_Mode 24/07 25 26 27

ch9_APS0: 00000004
  0000:  15 00 ff ff
ch9_NTSC_APS1: 0000003A
  0000:  15 5b ff 18 0b ff 19 06  1f 1a 05 ff 1b 0d ff 28
  0010:  00 ff 29 0d ff 2a 00 0f  2b 00 3f 2c 00 ff 2d 00
  0020:  ff 2e 15 3f 2f 15 3f 40  0f ff 41 0f ff 42 0f 7f
  0030:  43 f0 ff 44 00 7f 45 00  ff ff
ch9_NTSC_APS2: 0000003A
  0000:  15 5f ff 18 0b ff 19 06  1f 1a 05 ff 1b 0d ff 28
  0010:  ff ff 29 47 ff 2a 0d 0f  2b 11 3f 2c 9e ff 2d 2d
  0020:  ff 2e 11 3f 2f 11 3f 40  0f ff 41 0f ff 42 0f 7f
  0030:  43 f0 ff 44 00 7f 45 00  ff ff
ch9_NTSC_APS3: 0000003A
  0000:  15 5f ff 18 0b ff 19 06  1f 1a 05 ff 1b 0d ff 28
  0010:  ff ff 29 87 ff 2a 0b 0f  2b 15 3f 2c 98 ff 2d 29
  0020:  ff 2e 15 3f 2f 15 3f 40  0f ff 41 0f ff 42 0f 7f
  0030:  43 f0 ff 44 00 7f 45 00  ff ff
ch9_NTSC_Mode11: 0000001F
  0000:  16 25 7f 17 3b ff 38 5a  7f 39 0f 0f 3a 05 ff 3b
  0010:  05 ff 3c 41 7f 3d 41 7f  3e 05 ff 3f 05 ff ff
ch9_NTSC_Mode11_APS1: 00000001
  0000:  ff
ch9_NTSC_Mode11_APS2and3: 0000000D
  0000:  24 00 07 25 b4 ff 26 b4  ff 27 df ff ff
ch9_NTSC_Mode17: 0000001F
  0000:  16 25 7f 17 38 ff 38 56  7f 39 00 0f 3a fa ff 3b
  0010:  fa ff 3c 3e 7f 3d 3e 7f  3e fa ff 3f fa ff ff
ch9_NTSC_Mode17_APS2and3: 0000000D
  0000:  24 00 07 25 ad ff 26 ad  ff 27 d6 ff ff
ch9_NTSC_Mode29: 0000001F
  0000:  16 25 7f 17 48 ff 38 6d  7f 39 0f 0f 3a 3e ff 3b
  0010:  3e ff 3c 4f 7f 3d 4f 7f  3e 3e ff 3f 3e ff ff
ch9_NTSC_Mode29_APS2and3: 0000000D
  0000:  24 01 07 25 dd ff 26 dd  ff 27 11 ff ff
ch9_NTSC_Mode35: 0000001F
  0000:  16 25 7f 17 42 ff 38 64  7f 39 0f 0f 3a 23 ff 3b
  0010:  23 ff 3c 48 7f 3d 48 7f  3e 23 ff 3f 23 ff ff
ch9_NTSC_Mode35_APS2and3: 0000000D
  0000:  24 00 07 25 c9 ff 26 c9  ff 27 f9 ff ff
ch9_PAL_APS1: 0000003A
  0000:  15 1b ff 18 6c ff 19 03  1f 1a 33 ff 1b 38 ff 28
  0010:  00 ff 29 40 ff 2a 08 0f  2b 22 3f 2c 9b ff 2d 64
  0020:  ff 2e 22 3f 2f 22 3f 40  fe ff 41 7e ff 42 3f 7f
  0030:  43 e0 ff 44 15 7f 45 40  ff ff
ch9_PAL_Mode8: 0000002B
  0000:  16 25 7f 17 38 ff 24 00  07 25 bd ff 26 bd ff 27
  0010:  dd ff 38 4d 7f 39 0c 0f  3a 3c ff 3b 1b ff 3c 2d
  0020:  7f 3d 31 7f 3e 95 ff 3f  b7 ff ff
ch9_PAL_Mode14: 0000002B
  0000:  16 25 7f 17 3b ff 24 00  07 25 c7 ff 26 c7 ff 27
  0010:  e9 ff 38 51 7f 39 0c 0f  3a 4c ff 3b 29 ff 3c 30
  0020:  7f 3d 34 7f 3e 9d ff 3f  c0 ff ff
ch9_PAL_Mode26: 0000002B
  0000:  16 25 7f 17 48 ff 24 01  07 25 f2 ff 26 f2 ff 27
  0010:  1c ff 38 63 7f 39 0c 0f  3a 94 ff 3b 6a ff 3c 3a
  0020:  7f 3d 3f 7f 3e bf ff 3f  ea ff ff
ch9_PAL_Mode32: 0000002B
  0000:  16 25 7f 17 46 ff 24 01  07 25 ed ff 26 ed ff 27
  0010:  15 ff 38 60 7f 39 0c 0f  3a 89 ff 3b 60 ff 3c 38
  0020:  7f 3d 3e 7f 3e ba ff 3f  e3 ff ff 00 00 00

*/

/* Naming scheme:

ch    for all chrontel
ch_a  for 7007 family
ch_b  for 7009 family

cx
cx_a  for bt family
cx_b  for cx family

ph
ph_a
ph_b

numbering in TVChip:

<company><family><major><minor>

conexant bt868    2101
conexant bt869    2102
conexant cx25870  2201
conexant cx25871  2202
conexant cx25872  2203
conexant cx25873  2204
conexant cx25874  2205
conexant cx25875  2206

chrontel ch7003  1103
chrontel ch7004  1104
chrontel ch7005  1105
chrontel ch7006  1106
chrontel ch7007  1107
chrontel ch7008  1108

chrontel ch7009  1201
chrontel ch7010  1202
chrontel ch7011  1203

philips  saa7102/08   3102
philips  saa7103/09   3103
philips  saa7104/08A  3201
philips  saa7105/09A  3202

philips:

status: iferr. bferr
state: dclk, eidiv (but always set?)
pcle, pcli, pclsy, ifra, ifbp, fili, cvbsen2, 
yfil, xint, yupsc, 

extra filter setting??? (rename adaptive to extra filter)
can do upscaling modes, 640 huge, 1024 ..., 1280 ...

 */




